/* $Id: x11lib.c,v 1.36 2010/03/16 20:00:25 kiesling Exp $ -*-c-*-*/

/*
  This file is part of Ctalk.
  Copyright  2005-2010  Robert Kiesling, ctalk@ctalklang.org.
  Permission is granted to copy this software provided that this copyright
  notice is included in all source code modules.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software Foundation, 
  Inc., 51 Franklin St., Fifth Floor, Boston, MA 02110-1301 USA.
*/

/*
 *  Notes - This code should still be handled gently.  I've tried to
 *  avoid memory leaks, but it is possible to lose objects in the 
 *  X11TerminalStream methods, and probably elsewhere, if the code 
 *  is pushed.
 *
 *  Also, for Solaris 8, and probably Solaris 9, a program can
 *  generate a "too many open files" error.  The (possible - I haven't
 *  tried it) solution for this is the same as for servers like
 *  apache, mysql, and postgres - increase the system's maximum file
 *  descriptors from 256 to 65535.  Supposedly you can add lines
 *  similar to the following to the /etc/system file; for example: 
 *
 *    set rlim_fd_max=65535
 *
 *  Also see the ulimit manual page.  You should not need to worry
 *  about how much memory is in the data segment, however, because
 *  with this release, Ctalk sets the brk limit depending on how
 *  much memory the system can provide it.  And please let me know 
 *  if you see this message anyway.
 *
 *  Solaris 10 has several workarounds for this.  You should probably
 *  update if this is an issue.  Actually, a X application can run out 
 *  of file descriptors under any OS if it gets too large.
 *
 *  Linux is touchy about signal handling.  In particular, the default
 *  handlers do not seem to be reliably restartable.  If an application 
 *  receives a SIGUSR2 (though I'm not sure why the X libs should raise 
 *  this signal) or a SIGABRT, the only solution now is to kill the 
 *  application.  SA_RESETHAND seems to work best for all platforms.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <ctype.h>
#include "object.h"
#include "message.h"
#include "ctalk.h"
#include "rtinfo.h"
#include "x11defs.h"

#if ! defined (DJGPP) && ! defined (WITHOUT_X11)
#include <Xlib.h>
#include <Xutil.h>
#include <Xatom.h>
#include <keysym.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <signal.h>
#include <sys/wait.h>
#include <sys/ucontext.h>
#include <sys/socket.h>
#include <sys/un.h>

#if defined(__GNUC__) && defined(__sparc__) && defined(__svr4__)
#include <sys/resource.h>
#endif

#define X11LIB_PRINT_COPIOUS_DEBUG_INFO 0
#define X11LIB_FRAME 0
#define MAX_RESIZE_RETRIES 3

Display *display = NULL;
static int screen, screen_depth;
static Window root;

char *fixed_font_name = "*-fixed-*-*-*-*-*-120-*";
static char **fixed_font_set;
Font fixed_font = 0;
static int n_fixed_fonts;

#if !X11LIB_FRAME
static int frame_pane_sock_fd;
#endif

char *shm_mem;
int mem_id;

Atom wm_delete_window;

int __ctalkX11IOErrorHandler (Display *);
int __ctalkX11ErrorHandler (Display *, XErrorEvent *);
int (*default_io_handler) (Display *);
int (*default_error_handler) (Display *, XErrorEvent *);
static OBJECT *__x11_pane_win_depth_value_object (OBJECT *);
OBJECT *__x11_pane_font_id_value_object (OBJECT *);
int __have_bitmap_buffers (OBJECT *);
static void request_client_shutdown (void) {
  shm_mem[SHM_SHUTDOWN] = 't';
}

static int client_shutdown_requested (char *s) {
  return (s[SHM_SHUTDOWN] == 't') ? TRUE : FALSE;
}

/*
 *  We should let the signal handlers in class Application
 *  handle signals.  So these functions should not be needed.
 */
// #if defined(__sparc__) && defined(__svr4__)
// void x11_sig_handler (int sig, siginfo_t *sip, ucontext_t *ucp) {
// #else /* Linux */ 
// void x11_sig_handler (int sig, siginfo_t *sip, void *ucp) {
// #endif
//  struct sigaction old_action;
//  sigemptyset (&(old_action.sa_mask));
//#if X11LIB_PRINT_COPIOUS_DEBUG_INFO  
//  fprintf (stderr, "x11_sig_handler: Caught signal %d\n", sig);
//#endif
// /*   XSync (display, TRUE); */
//  kill (getpid (), sig);
//}

/*
 *  Although we do *not* simply want to terminate the input client
 *  process, resetting the signal handlers seems to work best to avoid
 *  unchecked signal propagation on all platforms.
 */
// static void x11_sig_init (void) {
//  struct sigaction action, old_action;
//  action.sa_flags = SA_SIGINFO|SA_RESETHAND;
//  action.sa_sigaction = x11_sig_handler;
//  sigemptyset (&(action.sa_mask));
//  sigaction (SIGHUP, &action, &old_action);
//  /*
//   *  Some systems complain a lot if a call like XPending is
//   *  interrupted by the user, so just use the system's CTRL-C
//   *  handler.
//   */
///*   sigaction (SIGINT, &action, &old_action); */
//  sigaction (SIGQUIT, &action, &old_action);
//  sigaction (SIGABRT, &action, &old_action);
//  sigaction (SIGPIPE, &action, &old_action);
//  sigaction (SIGTERM, &action, &old_action);
//  sigaction (SIGUSR1, &action, &old_action);
//  sigaction (SIGUSR2, &action, &old_action);
//  sigaction (SIGCHLD, &action, &old_action);
//}

int lookup_color (XColor *color, char *name) {

  XColor exact_color;
  Colormap default_cmap;

  default_cmap = DefaultColormap (display, screen);
  
  if (XAllocNamedColor 
      (display, default_cmap, name, &exact_color, color)) {
    return SUCCESS;
  } else {
    return ERROR;
  }
}

int lookup_pixel (char *color) {
  XColor c;
  if (!lookup_color (&c, color))
    return c.pixel;
  else
    return BlackPixel (display, screen);
}

extern int __xlib_clear_rectangle_basic (Drawable, int, int, int, int, GC);
int __xlib_put_str (Drawable w, GC gc, char *s) {
  char *c1, *c2, *c3, *c4, *c5;
  int n;
  int font_ascent = 0, font_descent = 0;  /* Avoid warnings. */
  static Font c_font;
  static XFontStruct *fontinfo;
  static char c_font_name[MAXLABEL], *c_term;
  int x, y, width, font_id;

  if ((n = sscanf (s, "%d:%d:%d:%d:%s:", &x, &y, 
		   &width, &font_id, c_font_name))!=5)
    return ERROR;
  if ((c_term = strchr (c_font_name, ':')) != NULL)
    *c_term = '\0';
#if X11LIB_PRINT_COPIOUS_DEBUG_INFO
  fprintf (stderr, "font_id (2) = %d, c_font = %d, c_font_name = %s\n", 
	   font_id, (int)c_font, c_font_name);
#endif
  if (font_id != (int)c_font) {
    if (c_font != fixed_font) {
      if ((c_font = XLoadFont (display, c_font_name)) == 0) {
#ifndef WITHOUT_X11_WARNINGS
	printf ("__xlib_put_str: Could not load font %s.\n", c_font_name);
	printf ("__xlib_put_str: (To disable these messages, build Ctalk with\n");
	printf ("__xlib_put_str: the --without-x11-warnings option.)\n");
#endif
	c_font = fixed_font;
      }
    } else {
      c_font = fixed_font;
    }
    fontinfo = XQueryFont (display, c_font);
    font_ascent = fontinfo -> max_bounds.ascent;
    font_descent = fontinfo -> max_bounds.descent;
    XFreeFontInfo (NULL, fontinfo, 0);
    XSetFont (display, gc, c_font);
  }
  c1 = strchr (s, ':'); ++c1; c2 = strchr (c1, ':'); ++c2;
  c3 = strchr (c2, ':'); ++c3; c4 = strchr (c3, ':'); ++c4;
  c5 = strchr (c4, ':'); ++c5;
  if (!*c5) 
    return ERROR;
  __xlib_clear_rectangle_basic (w, x, y - font_ascent, 
				width, 
				font_ascent + font_descent, gc);
  /*
   *  VT100 erase line - has the effect of clearing the 
   *  line.
   */
  if (strcmp (c5, "\x1b" "[2K"))
    XDrawString (display, w, gc, x, y, c5, strlen (c5));
  return SUCCESS;
}

int __xlib_set_wm_name_prop (Drawable drawable, GC gc, char *s) {
  XTextProperty text_prop, text_prop_return;
  Atom wm_name;
  char *s_1;
  if ((wm_name = XInternAtom (display, "WM_NAME", True)) == None) {
#ifndef WITHOUT_X11_WARNINGS
    printf ("__xlib_set_wm_name_prop: WM_NAME property not found.\n");
    printf ("__xlib_put_str: (To disable these messages, build Ctalk with\n");
    printf ("__xlib_put_str: the --without-x11-warnings option.)\n");
#endif
    return ERROR;
  }  else {
    s_1 = strdup (s);
    XStringListToTextProperty (&s_1, 1, &text_prop);
    text_prop_return.value = NULL;
    do {
      XSetTextProperty (display, drawable, &text_prop, wm_name);
      if (text_prop_return.value)
	XFree (text_prop_return.value);
    } while (!XGetTextProperty (display, drawable, 
				&text_prop_return, wm_name));
    if (text_prop_return.value)
      XFree (text_prop_return.value);
    XFree (text_prop.value);
    free (s_1);
  }
  XFlush (display);
  return SUCCESS;
}

int __xlib_set_wm_name_prop_new (Drawable drawable, GC gc, char *s) {
  XTextProperty text_prop, text_prop_return;
  Atom wm_name;
  char *s_1;
  if ((wm_name = XInternAtom (display, "WM_NAME", True)) == None) {
#ifndef WITHOUT_X11_WARNINGS
    printf ("__xlib_set_wm_name_prop: WM_NAME property not found.\n");
    printf ("__xlib_set_wm_name_prop: (To disable these messages, build Ctalk with\n");
    printf ("__xlib_set_wm_name_prop: the --without-x11-warnings option.)\n");
#endif
    return ERROR;
  }  else {
    s_1 = strdup (s);
    XStringListToTextProperty (&s_1, 1, &text_prop);
    text_prop_return.value = NULL;
    do {
      XSetTextProperty (display, drawable, &text_prop, wm_name);
    } while (!XGetTextProperty (display, drawable, 
				&text_prop_return, wm_name));
    if (text_prop.value)
      XFree (text_prop_return.value);
    if (text_prop_return.value)
      XFree (text_prop_return.value);
    free (s_1);
  }
  return SUCCESS;
}

int __xlib_clear_window (Drawable drawable, GC gc) {
   XClearWindow (display, drawable);
   return SUCCESS;
}

int __xlib_clear_rectangle (Drawable drawable, GC gc, char *data) {
  int r, x, y, width, height;
  int panebuffer_id, panebackingstore_id;
  XGCValues gcv, old_gcv;
  if ((r = sscanf (data, "%d:%d:%d:%d:%d:%d:", 
		   &x, &y, &width, &height,
		   &panebuffer_id, &panebackingstore_id)) != 6) {
#ifndef WITHOUT_X11LIB_WARNINGS
    _warning 
      ("__xlib_clear_rectangle: bad data conversion: data = %s, r = %d.\n",
       data, r);
    _warning 
      ("__xlib_clear_rectangle: (To disable these messages, build Ctalk with\n");
    _warning 
      ("__xlib_clear_rectangle: the --without-x11-warnings option.)\n");
#endif
    return ERROR;
  }
  if (!panebuffer_id && !panebackingstore_id) {
    XClearArea (display, drawable, x, y, width, height, True);
  } else {
#define CLEAR_RECTANGLE_GCV_MASK (GCFunction|GCForeground|GCBackground|GCFillStyle)
    XGetGCValues (display, gc, 
 		  CLEAR_RECTANGLE_GCV_MASK, &old_gcv);
    gcv.background = old_gcv.background;
    gcv.foreground = old_gcv.background;
    gcv.fill_style = FillSolid;
    gcv.function = GXcopy;
    XChangeGC (display, gc, CLEAR_RECTANGLE_GCV_MASK, &gcv);
    XFillRectangle (display, (Drawable)panebuffer_id, gc,
 		    x, y, width, height);
    XChangeGC (display, gc, CLEAR_RECTANGLE_GCV_MASK, &old_gcv);
  }
  return SUCCESS;
}

#define POINT_GCV_MASK (GCFunction | GCForeground | GCFillStyle) 

int __xlib_draw_point (Drawable drawable_arg, GC gc, char *data) {
  XGCValues point_gcv, old_gcv;
  int center_x, center_y, pen_width;
  int panebuffer_id, panebackingstore_id, actual_drawable;
  int r;
  char colorname[MAXMSG];
  
  if ((r = sscanf (data, "%d:%d:%d:%d:%d:%s", &center_x, &center_y, 
		   &pen_width,
 		   &panebuffer_id,
 		   &panebackingstore_id,
 		   colorname)) != 6) {
#ifndef WITHOUT_X11LIB_WARNINGS
    _warning ("__xlib_draw_point: bad data conversion: data = %s, r = %d.\n",
 	      data, r);
    _warning ("__xlib_draw_point: (To disable these messages, build Ctalk with\n");
    _warning ("__xlib_draw_point: the --without-x11-warnings option.)\n");
#endif
    return ERROR;
  } 

  if (!panebuffer_id && !panebackingstore_id) {
    actual_drawable = drawable_arg;
  } else {
    actual_drawable = panebuffer_id;
  }
  XGetGCValues (display, gc, DEFAULT_GCV_MASK, &old_gcv);
  point_gcv.function = GXcopy;
  point_gcv.foreground = lookup_pixel (colorname);
  point_gcv.fill_style = FillSolid;
  XChangeGC (display, gc, POINT_GCV_MASK, &point_gcv);
  if (pen_width < 2) {
    XDrawPoint (display, actual_drawable, gc,
		center_x, center_y);
  } else {
    XFillArc (display, actual_drawable, gc,
	      center_x, center_y,
	      pen_width, pen_width,
	      0, 360 * 64);
  }
  XChangeGC (display, gc, DEFAULT_GCV_MASK, &old_gcv);
  return SUCCESS;
}

#define LINE_GCV_MASK (GCFunction|GCForeground|GCFillStyle|GCLineWidth) 
int __xlib_draw_line (Drawable drawable_arg, GC gc, char *data) {
  XGCValues line_gcv, old_gcv;
  int line_start_x, line_start_y, line_end_x, line_end_y, pen_width;
  int panebuffer_id, panebackingstore_id, actual_drawable;
  int r;
  char colorname[MAXMSG];
  if ((r = sscanf (data, "%d:%d:%d:%d:%d:%d:%d:%s", 
		   &line_start_x, 
		   &line_start_y, 
		   &line_end_x, 
		   &line_end_y, 
		   &pen_width, 
		   &panebuffer_id,
		   &panebackingstore_id,
		   colorname)) != 8) {
#ifndef WITHOUT_X11LIB_WARNINGS
    _warning ("__xlib_draw_line: bad data conversion: data = %s, r = %d.\n",
	      data, r);
    _warning ("__xlib_draw_line: (To disable these messages, build Ctalk with\n");
    _warning ("__xlib_draw_line: the --without-x11-warnings option.)\n");
#endif
    return ERROR;
  }
  if (!panebuffer_id && !panebackingstore_id) {
    actual_drawable = drawable_arg;
  } else {
    actual_drawable = panebuffer_id;
  }
  line_gcv.function = GXcopy;
  line_gcv.foreground = lookup_pixel (colorname);
  line_gcv.fill_style = FillSolid;
  line_gcv.line_width = pen_width;
  XGetGCValues (display, gc, DEFAULT_GCV_MASK, &old_gcv);
  XChangeGC (display, gc, LINE_GCV_MASK, &line_gcv);
  XDrawLine (display, actual_drawable, gc,
	     line_start_x,
	     line_end_x,
	     line_start_y,
	     line_end_y);
  XChangeGC (display, gc, DEFAULT_GCV_MASK, &old_gcv);
  return SUCCESS;
}

#define RECTANGLE_GCV_MASK (GCFunction|GCForeground|GCFillStyle|GCLineWidth) 
int __xlib_draw_rectangle (Drawable drawable_arg, GC gc, char *data) {
  XGCValues rectangle_gcv, old_gcv;
  int top_start_x, top_start_y, right_end_x, right_end_y, pen_width,
    fill;
  int panebuffer_id, panebackingstore_id, actual_drawable;
  int r;
  char colorname[MAXMSG];
  if ((r = sscanf (data, "%d:%d:%d:%d:%d:%d:%d:%d:%s", 
		   &top_start_x, 
		   &top_start_y, 
		   &right_end_x, 
		   &right_end_y, 
		   &pen_width, 
		   &fill, 
		   &panebuffer_id, 
		   &panebackingstore_id,
		   colorname)) != 9) {
#ifndef WITHOUT_X11LIB_WARNINGS
    _warning ("__xlib_draw_rectangle: bad data conversion: data = %s, r = %d.\n",
	      data, r);
    _warning ("__xlib_draw_rectangle: (To disable these messages, build Ctalk with\n");
    _warning ("__xlib_draw_rectangle: the --without-x11-warnings option.)\n");
#endif
    return ERROR;
  }

  if (!panebuffer_id && !panebackingstore_id) {
    actual_drawable = drawable_arg;
  } else {
    actual_drawable = panebuffer_id;
  }
  rectangle_gcv.function = GXcopy;
  rectangle_gcv.foreground = lookup_pixel (colorname);
  rectangle_gcv.fill_style = FillSolid;
  rectangle_gcv.line_width = pen_width;
  XGetGCValues (display, gc, DEFAULT_GCV_MASK, &old_gcv);
  XChangeGC (display, gc, RECTANGLE_GCV_MASK, &rectangle_gcv);

  if (fill) {
    XFillRectangle (display, actual_drawable, gc,
		    top_start_x, top_start_y,
		    right_end_x,right_end_y);
  } else {
    XDrawRectangle (display, actual_drawable, gc,
		    top_start_x, top_start_y, right_end_x,
		    right_end_y);
  }
  XChangeGC (display, gc, DEFAULT_GCV_MASK, &old_gcv);
  return SUCCESS;
}

extern int __xlib_clear_pixmap (Drawable, unsigned int, unsigned int,
				GC gc);
int __xlib_resize_window (Drawable drawable, GC gc, char *data) {
  unsigned int width, height;
  int depth, r;
  unsigned int old_paneBuffer_id, old_paneBackingStore_id;
  Pixmap new_paneBuffer_id, new_paneBackingStore_id;
  if ((r = sscanf (data, "%u:%u:%d:%u:%u", 
		   &width, &height, &depth, 
		   &old_paneBuffer_id,
		   &old_paneBackingStore_id)) != 5) {
    fprintf (stderr, "__xlib_resize_window (sscanf width, height): %s : r = %d.\n", strerror (errno), r);
    return ERROR;
  }
  XResizeWindow (display, drawable, width, height);
  if (old_paneBuffer_id && old_paneBackingStore_id) {
    new_paneBuffer_id =
      XCreatePixmap (display, drawable, width, height, depth);
    __xlib_clear_pixmap (new_paneBuffer_id, width, height, gc);
    new_paneBackingStore_id =
      XCreatePixmap (display, drawable, width, height, depth);
    __xlib_clear_pixmap (new_paneBackingStore_id, width, height, gc);
    XCopyArea (display, (Pixmap)old_paneBuffer_id, new_paneBuffer_id, gc, 
	       0, 0, 
	       width, height, 
	       0, 0);
    XFreePixmap (display, (Pixmap)old_paneBuffer_id);
    XFreePixmap (display, (Pixmap)old_paneBackingStore_id);
    sprintf (data, "%u %u", (unsigned int)new_paneBuffer_id, 
	     (unsigned int)new_paneBackingStore_id);
  } else {
#ifdef __APPLE__
    /*
     *  Apple's shared memory management makes checking for an 
     *  actual bufferless pane difficult.
     */
    new_paneBuffer_id =
      XCreatePixmap (display, drawable, width, height, depth);
    __xlib_clear_pixmap (new_paneBuffer_id, width, height, gc);
    new_paneBackingStore_id =
      XCreatePixmap (display, drawable, width, height, depth);
    __xlib_clear_pixmap (new_paneBackingStore_id, width, height, gc);
    sprintf (data, "%u %u", 
	     (unsigned int)new_paneBuffer_id, 
	     (unsigned int)new_paneBackingStore_id);
#else
    if (depth) {
      new_paneBuffer_id =
	XCreatePixmap (display, drawable, width, height, depth);
      __xlib_clear_pixmap (new_paneBuffer_id, width, height, gc);
     new_paneBackingStore_id =
       XCreatePixmap (display, drawable, width, height, depth);
     __xlib_clear_pixmap (new_paneBackingStore_id, width, height, gc);
     sprintf (data, "%u %u", 
	      (unsigned int)new_paneBuffer_id, 
	      (unsigned int)new_paneBackingStore_id);
    } else {
      sprintf (data, "%d %d", 0, 0);
    }
#endif
  }
  return SUCCESS;
}

int __xlib_move_window (Drawable drawable, GC gc, char *data) {
  int x, y, r;
  if ((r = sscanf (data, "%d:%d", &x, &y)) != 2) {
    fprintf (stderr, "__xlib_move_window (sscanf x, y, width, height): %s : r = %d.\n", strerror (errno), r);
    return ERROR;
  }
  XMoveWindow (display, drawable, x, y);
  return SUCCESS;
}

int __xlib_refresh_window (Drawable drawable, GC gc, char *data) {
  int r;
  unsigned int buffer_id;
  unsigned int width, height;
  if ((r = sscanf (data, "%u:%u:%u", 
		   &buffer_id, 
		   &width, &height)) != 3) {
    fprintf (stderr, "__xlib_refresh_window: %s : r = %d.\n", 
	     strerror (errno), r);
    return ERROR;
  }
  XCopyArea (display, (Drawable)buffer_id, drawable, gc, 
	     0, 0,
	     width, height,
	     0, 0);
  return SUCCESS;
}

int __xlib_handle_client_request (int mem_handle) {
  unsigned int w;
  int r;
  char *shm_mem_2;
  GC gc;

  if ((shm_mem_2 = get_shmem (mem_handle)) == NULL) return 0;
  if (! *shm_mem_2) goto x_event_cleanup;

  if ((r = sscanf (&shm_mem_2[SHM_DRAWABLE], "%u", &w))<= 0) {
#ifndef WITHOUT_X11LIB_WARNINGS
      _warning ("__xlib_handle_client_request (sscanf w): %s.\n",
		strerror (errno));
      _warning ("__xlib_handle_client_request: (To disable these messages, build Ctalk with\n");
      _warning ("__xlib_handle_client_request: the --without-x11-warnings option.)\n");
#endif
      goto x_event_cleanup;
  }
  if ((r = sscanf (&shm_mem_2[SHM_GC], "0x%p", &gc)) <= 0) {
#ifndef WITHOUT_X11LIB_WARNINGS
    _warning ("__xlib_handle_client_request (sscanf gc): %s: %s.\n",
	      strerror (errno), &shm_mem_2[SHM_GC]);
    _warning ("__xlib_handle_client_request: (To disable these messages, build Ctalk with\n");
    _warning ("__xlib_handle_client_request: the --without-x11-warnings option.)\n");
#endif
    goto x_event_cleanup;
  }
  XFlush (display);
  switch (shm_mem_2[SHM_REQ]) 
    {
    case PANE_PUT_STR_REQUEST:
      __xlib_put_str ((Drawable)w, gc, &shm_mem_2[SHM_DATA]);
      break;
    case PANE_WM_TITLE_REQUEST:
      __xlib_set_wm_name_prop ((Drawable)w, gc, &shm_mem_2[SHM_DATA]);
      break;
    case PANE_DRAW_POINT_REQUEST:
      __xlib_draw_point ((Drawable)w, gc, &shm_mem_2[SHM_DATA]);
      break;
    case PANE_DRAW_LINE_REQUEST:
      __xlib_draw_line ((Drawable)w, gc, &shm_mem_2[SHM_DATA]);
      break;
    case PANE_DRAW_RECTANGLE_REQUEST:
      __xlib_draw_rectangle ((Drawable)w, gc, &shm_mem_2[SHM_DATA]);
      break;
    case PANE_CLEAR_WINDOW_REQUEST:
      __xlib_clear_window ((Drawable)w, gc);
      break;
    case PANE_CLEAR_RECTANGLE_REQUEST:
      __xlib_clear_rectangle ((Drawable)w, gc, &shm_mem_2[SHM_DATA]);
      break;
    case PANE_RESIZE_REQUEST:
      __xlib_resize_window ((Drawable)w, gc, &shm_mem_2[SHM_DATA]);
      strcpy (&shm_mem_2[SHM_RETVAL], &shm_mem_2[SHM_DATA]);
      break;
    case PANE_MOVE_REQUEST:
      __xlib_move_window ((Drawable)w, gc, &shm_mem_2[SHM_DATA]);
      break;
    case PANE_REFRESH_REQUEST:
      __xlib_refresh_window ((Drawable)w, gc, &shm_mem_2[SHM_DATA]);
      break;
    } 
 x_event_cleanup:
  shm_mem_2[SHM_REQ] = '\0';
  detach_shmem (mem_handle, shm_mem_2);
  return SUCCESS;
}

static char *__x11_pane_socket_path (int pid) {
  static char path[FILENAME_MAX];
  sprintf (path, "%s/ctalk.x.%d", P_tmpdir, pid);
  return path;
}

static int __x11_pane_stream_socket_open (void) {
  int fd;
  struct sockaddr_un addr;
  char path[FILENAME_MAX];
  if ((fd = socket (AF_UNIX, SOCK_DGRAM, 0)) == -1) {
    fprintf (stderr, "__x11_stream_socket: %s.\n", 
	     strerror (errno));
    return ERROR;
  }
  strcpy (path, __x11_pane_socket_path ((int)getpid ()));
  strncpy (addr.sun_path, path, sizeof (addr.sun_path) - 1);
  addr.sun_family = AF_UNIX;
  if (bind (fd, (struct sockaddr *) &addr, 
	    sizeof (struct sockaddr_un))) {
    fprintf (stderr, "__x11_stream_bind: %s. Exiting.\n", 
	     strerror (errno));
    exit (EXIT_FAILURE);
  }
  return fd;
}

static int __x11_pane_stream_socket_ready (int pid) {
  char path[FILENAME_MAX];
  struct stat statbuf;
  int r;
  strcpy (path, __x11_pane_socket_path (pid));
  /*
   *  If not open, wait half a sec and check again.
   */
  if ((r = stat (path, &statbuf)) == ERROR) {
    usleep (500000L);
    if ((r = stat (path, &statbuf)) == ERROR) {
      return ERROR;
    }
  } else {
    return SUCCESS;
  }
  return ERROR;
}

static int __x11_pane_socket_connect (int sock_fd) {
  int r;
  char path[FILENAME_MAX];
  struct sockaddr_un addr;
  strcpy (path, __x11_pane_socket_path ((int)getppid ()));
  strncpy (addr.sun_path, path, sizeof (addr.sun_path) - 1);
  addr.sun_family = AF_UNIX;
  if ((r = connect (sock_fd, 
	       (struct sockaddr *)&addr, 
		    sizeof (struct sockaddr_un))) == ERROR)
    fprintf (stderr, "__x11_pane_socket_connect: %s: %s\n", 
	     path, strerror (errno));
  return r;
}

static Display *__x11_open_display (void) {
  char *d_env = NULL;

  if (display != NULL) 
    return display;

  if ((display = XOpenDisplay (d_env)) != NULL) {
    default_io_handler = XSetIOErrorHandler (__ctalkX11IOErrorHandler);
    default_error_handler = XSetErrorHandler (__ctalkX11ErrorHandler);
    /* 
     *  See the comments above.
     */
    //     x11_sig_init ();
    screen = DefaultScreen (display);
    root = RootWindow (display, screen);
    screen_depth = DefaultDepth (display, screen);
    fixed_font_set = XListFonts (display, fixed_font_name, 10,
 				 &n_fixed_fonts);
    if (n_fixed_fonts) {
      fixed_font = XLoadFont (display, fixed_font_set[0]);
      XFreeFontNames (fixed_font_set);
    } else {
#ifndef WITHOUT_X11LIB_WARNINGS
      _warning ("__x11_open_display: Couldn't find fixed font.\n");
      _warning ("__x11_open_display: (To disable these messages, build Ctalk with\n");
      _warning ("__x11_open_display: the --without-x11-warnings option.)\n");
#endif
    }
    return display;
  }
  return NULL;
}

static void release_fixed_font (void) {
  if (n_fixed_fonts) { 
    /*
     *  This should not be needed on most displays.
     */
/*     XUnloadFont (display, fixed_font); */
/*     XFreeFontNames (fixed_font_set); */
    fixed_font = 0;
    n_fixed_fonts = 0;
  }
}

static Font get_user_font (OBJECT *self) {
  OBJECT *self_object, *fontDesc_var, *fontDesc_value_var;
  Font user_font;
  int n_user_fonts;
  char **user_font_set;
  self_object = IS_VALUE_INSTANCE_VAR(self) ? self -> __o_p_obj :
    self;
  if ((fontDesc_var = 
       __ctalkGetInstanceVariable (self_object, "fontDesc", TRUE)) 
      == NULL)
    return 0;
  if ((fontDesc_value_var = 
       __ctalkGetInstanceVariable (fontDesc_var, "value", TRUE)) 
      == NULL)
    return 0;
  if (!strcmp (fontDesc_value_var -> __o_value, NULLSTR))
    return 0;
    user_font_set = XListFonts (display, fontDesc_value_var->__o_value, 10,
 				 &n_user_fonts);
    if (n_user_fonts) {
      user_font = XLoadFont (display, user_font_set[0]);
      XFreeFontNames (user_font_set);
      return user_font;
    } else {
#ifndef WITHOUT_X11_WARNINGS
      _warning ("__x11_open_display: Couldn't find font %s.\n",
		fontDesc_value_var->__o_value);
      _warning ("__x11_open_display: (To disable these messages, build Ctalk with\n");
      _warning ("__x11_open_display: the --without-x11-warnings option.)\n");
#endif
    }
  return 0;
}

static void release_pane_gc (OBJECT *self) {
  OBJECT *gc_value_var;
  GC gc;
  gc_value_var = __x11_pane_win_gc_value_object (self);
  gc = (GC)generic_ptr_str (gc_value_var->__o_value);
  XFreeGC (display, gc);
  gc_value_var->__o_value[0] = '\0'; 
}

static void destroy_pane_window (OBJECT * self) {
  OBJECT *win_id_value_var;
  Window w;
  win_id_value_var = __x11_pane_win_id_value_object (self);
  w = atoi (win_id_value_var -> __o_value);
  XDestroyWindow (display, w);
}

#if X11LIB_FRAME

int __ctalkX11InputClient (OBJECT *streamobject, int parent_fd, int mem_handle, int main_win_id) {
  XEvent e;
  int events_waiting, handle_count;
  char *s;

  __x11_pane_socket_connect (parent_fd);

  while (1) {
    while ((events_waiting = XPending (display)) > 0) {
      XNextEvent (display, &e);
    }
    if ((s = (char *)get_shmem (mem_handle)) == NULL)
      return ERROR;
    if (client_shutdown_requested (s)) {
#if X11LIB_PRINT_COPIOUS_DEBUG_INFO  
      fprintf (stderr, "Process %d releasing shared memory.\n",
	       (int) getpid ());
#endif
      handle_count = detach_shmem (mem_id, s);
#if X11LIB_PRINT_COPIOUS_DEBUG_INFO  
      fprintf (stderr, "Shared memory attached to %d processes.\n",
	       handle_count);
#endif
      return SUCCESS;
    } else {
      handle_count = detach_shmem (mem_id, s);
    }
  }
  return SUCCESS;
}
#endif /* X11LIB_FRAME */

#if !X11LIB_FRAME
static OBJECT *__stream_queue_object (OBJECT *streamobject) {
  OBJECT *p_obj;
  static OBJECT *queue_object;
  if (!strcmp (streamobject -> __o_name, "value") &&
      streamobject -> __o_p_obj) {
    p_obj = streamobject -> __o_p_obj;
  } else {
    p_obj = streamobject;
  }
  if ((queue_object = 
       __ctalkGetInstanceVariable (p_obj, "inputQueue", FALSE))
      == NULL)
    return NULL;
  return queue_object;
}
#endif

#if !X11LIB_FRAME
static OBJECT *__x11_stream_pid_value_object (OBJECT *streamobject) {
  static OBJECT *pid_object,
    *pid_value_object;
  OBJECT *p_obj;
  if (!strcmp (streamobject -> __o_name, "value") &&
      streamobject -> __o_p_obj) {
    p_obj = streamobject -> __o_p_obj;
  } else {
    p_obj = streamobject;
  }
  if ((pid_object = 
       __ctalkGetInstanceVariable (p_obj, "inputPID", FALSE))
      == NULL)
    return NULL;
  if ((pid_value_object = 
       __ctalkGetInstanceVariable (pid_object, "value", FALSE))
      == NULL)
    return NULL;
  return pid_value_object;
}

static OBJECT *__x11_stream_client_sock_fd_value_object (OBJECT *streamobject) {
  static OBJECT *sock_fd_object,
    *client_sock_fd_value_object;
  OBJECT *p_obj;
  if (!strcmp (streamobject -> __o_name, "value") &&
      streamobject -> __o_p_obj) {
    p_obj = streamobject -> __o_p_obj;
  } else {
    p_obj = streamobject;
  }
  if ((sock_fd_object = 
       __ctalkGetInstanceVariable (p_obj, "clientFD", FALSE))
      == NULL)
    return NULL;
  if ((client_sock_fd_value_object = 
       __ctalkGetInstanceVariable (sock_fd_object, "value", FALSE))
      == NULL)
    return NULL;
  return client_sock_fd_value_object;
}
#endif

int __ctalkOpenX11InputClient (OBJECT *streamobject) {
#if !X11LIB_FRAME
  OBJECT *pid_value_object, *client_sock_fd_value_object;
  OBJECT *x11_pane_stream_object, *x11_pane_stream_queue_list;
  char buf[MAXLABEL];
  int main_win_id;
  OBJECT *pane_win_id_value_object;
#endif
  int child_pid/* , pane_sock_fd */;

  if ((display = __x11_open_display ()) == NULL)
    return -1;

  if ((mem_id = mem_setup ()) == ERROR) 
    return ERROR;

#if !X11LIB_FRAME
  x11_pane_stream_object = streamobject;
  if ((x11_pane_stream_queue_list = __stream_queue_object (streamobject))
      == NULL) {
    x11_pane_stream_object = NULL;
    release_fixed_font ();
    /*
     *  Delete GC and window here also, *if any*.
     */
    XCloseDisplay (display);
    display = NULL;
    return -3;
  }
#endif

  pane_win_id_value_object = 
    __x11_pane_win_id_value_object (streamobject->__o_p_obj);
  main_win_id = atoi (pane_win_id_value_object -> __o_value);

  /*
   *  *** Handle this error in the method.
   */
  if ((frame_pane_sock_fd = __x11_pane_stream_socket_open ()) == ERROR)
    return ERROR;

  switch (child_pid = fork ())
    {
    case 0:
#if X11LIB_PRINT_COPIOUS_DEBUG_INFO  
      fprintf (stderr, "Process %d started.  Parent PID %d.\n",
	       (int)getpid(), (int)getppid ());
#endif
      __ctalkX11InputClient (streamobject, frame_pane_sock_fd, mem_id, 
			     main_win_id);
#if X11LIB_PRINT_COPIOUS_DEBUG_INFO  
      fprintf (stderr, "Process %d exit.\n", (int)getpid ());
#endif
      _exit (0);
      break;
    case -1:
      fprintf (stderr, "__ctalkOpenX11InputClient: %s\n", 
	       strerror (errno));
      _exit (1);
      break;
    default:
      if (__x11_pane_stream_socket_ready (getpid ())) {
	fprintf (stderr, "__ctalkOpenX11InputClient: socket not ready.\n");
	return ERROR;
      }
      if ((shm_mem = (char *)get_shmem (mem_id)) == NULL)
	_exit (0);
#if !X11LIB_FRAME
      if ((pid_value_object = __x11_stream_pid_value_object (streamobject))
	  != NULL) {
	sprintf (buf, "%d", child_pid);
	__ctalkSetObjectValue (pid_value_object, buf);
      }
      if ((client_sock_fd_value_object = 
	   __x11_stream_client_sock_fd_value_object (streamobject)) != NULL) {
	sprintf (buf, "%d", frame_pane_sock_fd);
	__ctalkSetObjectValue (client_sock_fd_value_object, buf);
      }
#endif
      break;
    }
  return SUCCESS;
}

/*
 *  We have to be careful here and in __ctalkX11CloseParentPane 
 *  and __ctalkX11CloseClient that we don't destroy objects 
 *  asynchronously.
 */
int __ctalkCloseX11Pane (OBJECT *self) {
  OBJECT *gc_value_var;
  GC gc;
  OBJECT *win_id_value_var;
  Window w;

  gc_value_var = __x11_pane_win_gc_value_object (self);

  if ((gc = (GC)generic_ptr_str (gc_value_var->__o_value)) != NULL) {
    XFreeGC (display, gc);
    gc_value_var->__o_value[0] = '\0'; 
  }

/*   destroy_pane_window (self); */
  win_id_value_var = __x11_pane_win_id_value_object (self);
  if ((w = atoi (win_id_value_var -> __o_value)) != (Window)0) 
    XDestroyWindow (display, w);

  if (__have_bitmap_buffers (self)) {
    __ctalkX11FreePaneBuffer (self);
  }  

/*   request_client_shutdown (); */
/*   display = NULL; */
  return SUCCESS;
}

int __ctalkX11CloseParentPane (OBJECT *self) {
  OBJECT *pane_win_id_value_object;
  int main_win_id;
  pane_win_id_value_object = 
    __x11_pane_win_id_value_object (self);
  if ((main_win_id = atoi (pane_win_id_value_object -> __o_value))
      == 0)
    return ERROR;
  release_fixed_font ();
  request_client_shutdown ();
  display = NULL;
  return SUCCESS;
}

int __ctalkX11CloseClient (OBJECT *self) {
  int waitstatus;
#if X11LIB_PRINT_COPIOUS_DEBUG_INFO  
  fprintf (stderr, "__ctalkX11CloseClient: Waiting for client shutdown....\n");
#endif
  request_client_shutdown ();
  wait (&waitstatus);
#if X11LIB_PRINT_COPIOUS_DEBUG_INFO  
  fprintf (stderr, "...waited.\n");
#endif
#if X11LIB_PRINT_COPIOUS_DEBUG_INFO  
      fprintf (stderr, "App releasing shared memory.\n");
#endif
      release_shmem (mem_id, shm_mem);
#if X11LIB_PRINT_COPIOUS_DEBUG_INFO  
      fprintf (stderr, "Shared memory released.\n");
#endif
  delete_shmem_info (mem_id);
#if !X11LIB_FRAME
  if (shutdown (frame_pane_sock_fd, SHUT_RD))
    fprintf (stderr, "__ctalkX11CloseClient: shutdown: %s.\n",
	     strerror (errno));
  if (unlink (__x11_pane_socket_path ((int)getpid())))
    fprintf (stderr, "__ctalkX11CloseClient: unlink: %s : %s.\n",
	     __x11_pane_socket_path ((int)getpid()), strerror (errno));
#endif

  return SUCCESS;
}


#if X11LIB_FRAME
int __ctalkCreateX11MainWindow (OBJECT *self_object) {
   return SUCCESS;
}
int __ctalkCreateX11SubWindow (OBJECT *parent, OBJECT *self) {
  return SUCCESS;
}
int __ctalkMapX11Window (OBJECT *self_object) {
  return SUCCESS;
}
int __ctalkRaiseX11Window (OBJECT *self_object) {
  return SUCCESS; 
}
int __ctalkX11SetWMNameProp (OBJECT *self_object, char *title) {
  return SUCCESS;
}
int __ctalkX11UseFont (OBJECT *self) {
  return SUCCESS;
}
int __ctalkX11ResizeWindow (OBJECT *self, int width, int height) {
  return SUCCESS;
}
int __ctalkX11MoveWindow (OBJECT *self, int x, int y) {
  return SUCCESS;
}
#endif /* X11LIB_FRAME */

#if !X11LIB_FRAME
/*
 *  Only tests paneBuffer variable.  Might need to test
 *  paneBackingStore variable also if the buffers are not set 
 *  as a unit.
 */
int __have_bitmap_buffers (OBJECT *self_object) {
  OBJECT *paneBuffer_var;
  if ((paneBuffer_var =
       __ctalkGetInstanceVariable (self_object, "paneBuffer", TRUE))
      == NULL)
    return ERROR;
  if ((!strcmp (paneBuffer_var->instancevars->__o_classname, "Bitmap")) ||
      (__ctalkIsSubClassOf (paneBuffer_var->instancevars->__o_classname,
			    "Bitmap") > 0)) {
    if (strcmp (paneBuffer_var->instancevars->__o_value, "0") &&
	strcmp (paneBuffer_var->instancevars->__o_value, NULLSTR))
      return TRUE;
  }
  return FALSE;
}

int __ctalkX11InputClient (OBJECT *streamobject, int parent_fd, int mem_handle, int main_win_id) {
  XEvent e;
  int events_waiting, handle_count;
  int eventclass, eventdata1, eventdata2, eventdata3, eventdata4,
    eventdata5;
  char *s;
  char buf[MAXMSG];

  eventclass = eventdata1 = eventdata2 = eventdata3 = eventdata4 = 
    eventdata5 = 0;

  __x11_pane_socket_connect (parent_fd);

  while (1) {
    __xlib_handle_client_request (mem_handle);
    while ((events_waiting = XPending (display)) > 0) {
      __xlib_handle_client_request (mem_handle);
      XNextEvent (display, &e);
      switch (e.type)
	{
	case ButtonPress:
	  eventclass = BUTTONPRESS;
	  eventdata1 = e.xbutton.x;
	  eventdata2 = e.xbutton.y;
	  eventdata3 = e.xbutton.state;
	  eventdata4 = e.xbutton.button;
	  eventdata5 = 0;
	  break;
	case ButtonRelease:
	  eventclass = BUTTONRELEASE;
	  eventdata1 = e.xbutton.x;
	  eventdata2 = e.xbutton.y;
	  eventdata3 = e.xbutton.state;
	  eventdata4 = e.xbutton.button;
	  eventdata5 = 0;
	  break;
	case KeyPress:
	  eventclass = KEYPRESS;
	  eventdata1 = e.xkey.x;
	  eventdata2 = e.xkey.y;
	  eventdata3 = e.xkey.state;
	  eventdata4 = e.xkey.keycode;
	  eventdata5 = 
	    get_x11_keysym (e.xkey.keycode, e.xkey.state, TRUE);
	  break;
	case KeyRelease:
	  eventclass = KEYRELEASE;
	  eventdata1 = e.xkey.x;
	  eventdata2 = e.xkey.y;
	  eventdata3 = e.xkey.state;
	  eventdata4 = e.xkey.keycode;
	  eventdata5 = 
	    get_x11_keysym (e.xkey.keycode, e.xkey.state, FALSE);
	  break;
  	case MotionNotify:
  	  eventclass = MOTIONNOTIFY;
	  eventdata1 = e.xmotion.x;
	  eventdata2 = e.xmotion.y;
	  eventdata3 = e.xmotion.state;
	  eventdata4 = e.xmotion.is_hint;
	  eventdata5 = 0;
  	  break;
  	case MapNotify:
  	  eventclass = MAPNOTIFY;
	  eventdata1 = e.xmap.event;
	  eventdata2 = e.xmap.window;
	  eventdata3 = eventdata4 = eventdata5 = 0;
  	  break;
   	case Expose:
  	  eventclass = EXPOSE;
  	  eventdata1 = e.xexpose.x;
  	  eventdata2 = e.xexpose.y;
  	  eventdata3 = e.xexpose.width;
  	  eventdata4 = e.xexpose.height;
  	  eventdata5 = e.xexpose.count;
   	  break;
	case ConfigureNotify:
	  if (e.xconfigure.window == main_win_id) {
	    if (e.xconfigure.x && e.xconfigure.y) {
	      eventclass = MOVENOTIFY;
	    } else {
	      eventclass = RESIZENOTIFY;
	    }
	    eventdata1 = e.xconfigure.x;
	    eventdata2 = e.xconfigure.y;
	    eventdata3 = e.xconfigure.width;
	    eventdata4 = e.xconfigure.height;
	    eventdata5 = e.xconfigure.border_width;
	  }
	  break;
	case ClientMessage:
/* #if defined(__sparc__) && defined(__svr4__) */
	  if(e.xclient.data.l[0] == wm_delete_window) {
	    eventclass = WINDELETE;
	    eventdata1 = eventdata2 = eventdata3 = eventdata4 = 
	      eventdata5 = 0;
	  }
/* #else */
/* 	  eventclass = eventdata1 = eventdata2 = eventdata3 = eventdata4 =  */
/* 	      eventdata5 = 0; */
/* #endif	   */
	  break;
	}
      if (eventclass) {
	sprintf (buf, "%d:%d:%d:%d:%d:%d", eventclass, eventdata1, eventdata2,
		 eventdata3, eventdata4, eventdata5);
	write (parent_fd, buf, strlen (buf));
	eventclass = eventdata1 = eventdata2 = 0;
      }
    }
    if ((s = (char *)get_shmem (mem_id)) == NULL)
      return ERROR;
    if (client_shutdown_requested (s)) {
#if X11LIB_PRINT_COPIOUS_DEBUG_INFO  
      fprintf (stderr, "Process %d releasing shared memory.\n",
	       (int) getpid ());
#endif
      handle_count = detach_shmem (mem_id, s);
#if X11LIB_PRINT_COPIOUS_DEBUG_INFO  
      fprintf (stderr, "Shared memory attached to %d processes.\n",
	       handle_count);
#endif
      if (shutdown (parent_fd, SHUT_WR))
	fprintf (stderr, "__ctalkX11InputClient: shutdown: %s.\n",
		 strerror (errno));
      return SUCCESS;
    } else {
      handle_count = detach_shmem (mem_id, s);
    }
  }
  return SUCCESS;
}

OBJECT *__x11_pane_win_id_value_object (OBJECT *paneobject) {
  static OBJECT *pane_self_object, 
    *winid_var_object,
    *winid_var_value_object;

  if (!strcmp (paneobject -> __o_name, "value") &&
      paneobject -> __o_p_obj)
    pane_self_object = paneobject -> __o_p_obj;
  else
    pane_self_object = paneobject;
  if ((winid_var_object = 
       __ctalkGetInstanceVariable (pane_self_object, "xWindowID", TRUE))
      == NULL)
    return NULL;
  if ((winid_var_value_object = 
       __ctalkGetInstanceVariable (winid_var_object, "value", TRUE))
      == NULL)
    return NULL;

  return winid_var_value_object;
}

OBJECT *__x11_pane_win_gc_value_object (OBJECT *paneobject) {
  static OBJECT *pane_self_object, 
    *wingc_var_object,
    *wingc_var_value_object;

  if (!strcmp (paneobject -> __o_name, "value") &&
      paneobject -> __o_p_obj)
    pane_self_object = paneobject -> __o_p_obj;
  else
    pane_self_object = paneobject;
  if ((wingc_var_object = 
       __ctalkGetInstanceVariable (pane_self_object, "xGC", TRUE))
      == NULL)
    return NULL;
  if ((wingc_var_value_object = 
       __ctalkGetInstanceVariable (wingc_var_object, "value", TRUE))
      == NULL)
    return NULL;

  return wingc_var_value_object;
}

static OBJECT *__x11_pane_win_depth_value_object (OBJECT *paneobject) {
  static OBJECT *pane_self_object, 
    *depth_var_object,
    *depth_var_value_object;

  if (!strcmp (paneobject -> __o_name, "value") &&
      paneobject -> __o_p_obj)
    pane_self_object = paneobject -> __o_p_obj;
  else
    pane_self_object = paneobject;
  if ((depth_var_object = 
       __ctalkGetInstanceVariable (pane_self_object, "depth", TRUE))
      == NULL)
    return NULL;
  if ((depth_var_value_object = 
       __ctalkGetInstanceVariable (depth_var_object, "value", TRUE))
      == NULL)
    return NULL;

  return depth_var_value_object;
}

OBJECT *__x11_pane_font_id_value_object (OBJECT *paneobject) {
  static OBJECT *pane_self_object, *pane_font_object,
    *pane_font_fontid_object, *pane_font_fontid_value_object; 
    
  pane_self_object = (IS_VALUE_INSTANCE_VAR(paneobject) ?
		      paneobject -> __o_p_obj : paneobject);
  if ((pane_font_object = 
       __ctalkGetInstanceVariable (pane_self_object, "font", TRUE))
      == NULL)
    return NULL;
  if ((pane_font_fontid_object = 
       __ctalkGetInstanceVariable (pane_font_object, "fontId", TRUE)) 
      == NULL)
    return NULL;
  if ((pane_font_fontid_value_object = 
       __ctalkGetInstanceVariable (pane_font_fontid_object,
				   "value", TRUE)) == NULL)
    return NULL;
  return pane_font_fontid_value_object;
}

int __ctalkCreateX11MainWindow (OBJECT *self_object) {
  Window win_id;
  OBJECT *win_id_value_obj, *win_gc_value_obj, *win_depth_value_obj;
  XSetWindowAttributes set_attributes;
  XGCValues gcv;
  GC gc;
  XWMHints wm_hints;
  char buf[MAXLABEL];
  static int wm_event_mask = WM_CONFIGURE_EVENTS | WM_INPUT_EVENTS;
  if (!__x11_open_display ())
    return ERROR;
  win_id_value_obj =
    __x11_pane_win_id_value_object (self_object);
  win_gc_value_obj = 
    __x11_pane_win_gc_value_object (self_object);
  win_depth_value_obj = 
    __x11_pane_win_depth_value_object (self_object);
  set_attributes.backing_store = Always;
  win_id = XCreateWindow (display, root, 
 			  __pane_x_org (self_object), 
			  __pane_y_org (self_object), 
 			  __pane_x_size (self_object),
 			  __pane_y_size (self_object), 
 			  1, screen_depth,
 			  CopyFromParent, CopyFromParent, 
 			  CWBackingStore, &set_attributes);
  wm_hints.flags = (InputHint|StateHint);
  wm_hints.input = False;
  wm_hints.initial_state = NormalState;
  XSetWMHints (display, win_id, &wm_hints);
  wm_delete_window = XInternAtom (display, "WM_DELETE_WINDOW", False);
  XSetWMProtocols (display, win_id, &wm_delete_window, 1);
  XSetWindowBorder (display, win_id, BlackPixel(display, screen));
  XSelectInput(display, win_id, wm_event_mask);
  gcv.fill_style = FillSolid;
  gcv.function = GXcopy;
  gcv.foreground = BlackPixel (display, screen);
  gcv.background = WhitePixel (display, screen);
  if ((gcv.font = get_user_font (self_object)) == 0) {
    if (n_fixed_fonts && !fixed_font)
      fixed_font = XLoadFont (display, fixed_font_set[0]);
    if (fixed_font)
      gcv.font = fixed_font;
  }
  gc = XCreateGC (display, win_id, DEFAULT_GCV_MASK, &gcv);
  XSetWindowBackground (display, win_id, WhitePixel (display, screen));
  __xlib_set_wm_name_prop 
    (win_id, gc, 
     basename_w_extent(__argvFileName ()));
  sprintf (buf, "%#x", (unsigned int)gc);
  __ctalkSetObjectValue (win_gc_value_obj, buf);
  sprintf (buf, "%d", (int)win_id);
  __ctalkSetObjectValue (win_id_value_obj, buf);
  sprintf (buf, "%d", DefaultDepth (display, DefaultScreen (display)));
  __ctalkSetObjectValue (win_depth_value_obj, buf);
  return win_id;
}

int __ctalkCreateX11SubWindow (OBJECT *parent, OBJECT *self) {
  OBJECT *parent_object, *self_object;
  OBJECT *parent_win_id_value_object, *self_win_id_value_object,
    *win_depth_value_obj;
  OBJECT *self_gc_value_object;
  Window parent_id, self_id;
  XWindowAttributes parent_attributes;
  XSetWindowAttributes self_attributes;
  XGCValues gcv;
  GC gc;
  char buf[MAXLABEL];
  self_object = IS_VALUE_INSTANCE_VAR(self) ? self -> __o_p_obj : self;
  parent_object = IS_VALUE_INSTANCE_VAR(parent) ? 
    parent -> __o_p_obj : parent;
  parent_win_id_value_object =
    __x11_pane_win_id_value_object (parent_object);
  self_win_id_value_object =
    __x11_pane_win_id_value_object (self_object);
  self_gc_value_object = 
    __x11_pane_win_gc_value_object (self_object);
  win_depth_value_obj =
    __x11_pane_win_depth_value_object (self_object);
  parent_id = (Window) atoi(parent_win_id_value_object -> __o_value);
  if (!display) {
    _warning ("__ctalkX11CreateSubWindow: Could not open display, \"%s.\"\n",
	      getenv ("DISPLAY"));
    exit (EXIT_FAILURE);
  }
  if (parent_id == 0) {
    _warning ("__ctalkX11CreateSubWindow: Invalid parent window.\n");
    return ERROR;
  }
  XGetWindowAttributes (display, parent_id, &parent_attributes);
  self_attributes.backing_store = parent_attributes.backing_store;
  self_id = XCreateWindow (display, parent_id, 
			   0, 0,
			   parent_attributes.width,
			   parent_attributes.height,
			   0, /* border_width */
			   CopyFromParent,
			   CopyFromParent, CopyFromParent, 
			   CWBackingStore, &self_attributes);
  XSelectInput(display, self_id, (WM_CONFIGURE_EVENTS|WM_INPUT_EVENTS));
  gcv.fill_style = FillSolid;
  gcv.function = GXcopy;
  gcv.foreground = BlackPixel (display, screen);
  gcv.background = WhitePixel (display, screen);
  if ((gcv.font = get_user_font (self_object)) == 0) {
    if (n_fixed_fonts && !fixed_font)
      fixed_font = XLoadFont (display, fixed_font_set[0]);
    if (fixed_font)
      gcv.font = fixed_font;
  }
  gc = XCreateGC (display, self_id, DEFAULT_GCV_MASK, &gcv);
  sprintf (buf, "%d", (int)self_id);
  __ctalkSetObjectValue (self_win_id_value_object, buf);
  sprintf (buf, "%#x", (unsigned int)gc);
  __ctalkSetObjectValue (self_gc_value_object, buf);
  sprintf (buf, "%d", DefaultDepth (display, DefaultScreen (display)));
  __ctalkSetObjectValue (win_depth_value_obj, buf);
  return SUCCESS;
}

int __ctalkMapX11Window (OBJECT *self_object) {
  unsigned int win_id;
  OBJECT *win_id_value_obj;
  win_id_value_obj =
    __x11_pane_win_id_value_object (self_object);
  sscanf (win_id_value_obj -> __o_value, "%d", &win_id);
  XMapWindow (display, (Window)win_id);
  return SUCCESS;
}

int __ctalkRaiseX11Window (OBJECT *self_object) {
  unsigned int win_id;
  OBJECT *win_id_value_obj;
  XWindowChanges config;
  win_id_value_obj =
    __x11_pane_win_id_value_object (self_object);
  sscanf (win_id_value_obj -> __o_value, "%d", &win_id);
  XRaiseWindow (display, (Window)win_id);
  config.x = __pane_x_org (self_object); 
  config.y = __pane_y_org (self_object);
  XConfigureWindow (display, win_id, CWX | CWY, &config);
  return SUCCESS; 
}

int __ctalkX11SetWMNameProp (OBJECT *self_object, char *title) {
  OBJECT *win_id_value, *gc_value;
  win_id_value = __x11_pane_win_id_value_object (self_object);
  gc_value = __x11_pane_win_gc_value_object (self_object);
  make_req (shm_mem, PANE_WM_TITLE_REQUEST,
	    win_id_value->__o_value,
	    gc_value->__o_value, title);
  return SUCCESS;
}

int __ctalkX11UseFont (OBJECT *self) {
  OBJECT *self_object, *gc_value, *font_id_value;
  GC gc;
  XGCValues gc_values;
  Font new_font;
  int r;
  char buf[MAXMSG];
  self_object = (IS_VALUE_INSTANCE_VAR (self) ? self->__o_p_obj : self);
  gc_value = __x11_pane_win_gc_value_object (self_object);
  font_id_value = __x11_pane_font_id_value_object (self_object);
  if (sscanf (gc_value -> __o_value, "0x%p", &gc) != 1) {
    fprintf (stderr, "__ctalkX11UseFont: Bad gc value.\n");
    return ERROR;
  }
  XGetGCValues (display, gc, DEFAULT_GCV_MASK, &gc_values);
  if (gc_values.font && gc_values.font != fixed_font) {
    XUnloadFont (display, gc_values.font);
  }
  if ((new_font = get_user_font (self_object)) != 0) {
    gc_values.font = new_font;
    sprintf (buf, "%d", (int) new_font);
    __ctalkSetObjectValue (font_id_value, buf);
    r = XChangeGC (display, gc, DEFAULT_GCV_MASK, &gc_values);
  }
  return SUCCESS;
}

static int __x11_resize_request_internal (int width, int height, int depth,
				      char *win_id_value,
				      char *gc_value,
				      char *paneBuffer_id_value,
				      char *paneBackingStore_id_value,
				      int *paneBuffer_id_return,
				      int *paneBackingStore_id_return) {
  char d_buf[MAXMSG], d_buf_2[MAXMSG];
  int r;
#ifdef GRAPHICS_WRITE_SEND_EVENT
  XEvent send_event;
#endif
  sprintf (d_buf, "%d:%d:%d:%s:%s", width, height, depth,
	   paneBuffer_id_value,
	   paneBackingStore_id_value);
  make_req (shm_mem, PANE_RESIZE_REQUEST,
	    win_id_value, gc_value, d_buf);
#ifdef GRAPHICS_WRITE_SEND_EVENT
  send_event.xgraphicsexpose.type = GraphicsExpose;
  send_event.xgraphicsexpose.send_event = True;
  send_event.xgraphicsexpose.display = display;
  send_event.xgraphicsexpose.drawable = atoi(win_id_value);
  XSendEvent (display, atoi(win_id_value), False, 0L, &send_event);
#endif
  wait_req (shm_mem);
  sscanf (&shm_mem[SHM_RETVAL], "%s %s", d_buf, d_buf_2);
  r = chkatoi (d_buf);
  if (errno) {
#ifndef WITHOUT_X11LIB_WARNINGS
    _warning ("__x11_resize_request_internal: Invalid new buffers.\n");
    _warning ("__x11_resize_request_internal: (To disable these messages, build Ctalk with\n");
    _warning ("__x11_resize_request_internal: the --without-x11-warnings option.)\n");
#endif
    return ERROR;
  } else {
    *paneBuffer_id_return = r;
  }
  r = chkatoi (d_buf_2);
  if (errno) {
#ifndef WITHOUT_X11LIB_WARNINGS
    _warning ("__x11_resize_request_internal: Invalid new buffers.\n");
    _warning ("__x11_resize_request_internal: (To disable these messages, build Ctalk with\n");
    _warning ("__x11_resize_request_internal: the --without-x11-warnings option.)\n");
#endif
    return ERROR;
  } else {
    *paneBackingStore_id_return = r;
  }
  return SUCCESS;
}

int __ctalkX11ResizeWindow (OBJECT *self, int width, int height,
			    int depth) {
  OBJECT *self_object, *win_id_value, *gc_value;
  OBJECT *paneBuffer_object, *paneBackingStore_object;
  OBJECT *pane_size_point_object, *pane_size_point_y_object, 
    *pane_size_point_x_object;
  char d_buf[MAXLABEL];
  int new_paneBuffer_id, new_paneBackingStore_id, n_resize_retries;
#ifdef GRAPHICS_WRITE_SEND_EVENT
  XEvent send_event;
#endif
  self_object = (IS_VALUE_INSTANCE_VAR (self) ? self->__o_p_obj : self);
  win_id_value = __x11_pane_win_id_value_object (self_object);
  gc_value = __x11_pane_win_gc_value_object (self_object);
  if (__have_bitmap_buffers (self_object)) {
    if ((pane_size_point_object =
	 __ctalkGetInstanceVariable (self_object, "size", TRUE))
	== NULL)
      return ERROR;
    if ((pane_size_point_y_object =
	 __ctalkGetInstanceVariable (pane_size_point_object, "y", TRUE))
	== NULL)
      return ERROR;
    if ((pane_size_point_x_object =
	 __ctalkGetInstanceVariable (pane_size_point_object, "x", TRUE))
	== NULL)
      return ERROR;
    if ((paneBuffer_object =
	 __ctalkGetInstanceVariable (self_object, "paneBuffer", TRUE))
	== NULL) 
      return ERROR;
    if ((paneBackingStore_object =
	 __ctalkGetInstanceVariable (self_object, "paneBackingStore", TRUE))
	== NULL) 
      return ERROR;
    n_resize_retries = 0;
  try_resize_request:
    if (!__x11_resize_request_internal 
	(width, height, depth,
	 win_id_value -> __o_value,
	 gc_value -> __o_value,
	 paneBuffer_object->instancevars->__o_value,
	 paneBackingStore_object->instancevars->__o_value,
	 &new_paneBuffer_id, &new_paneBackingStore_id)) {
      sprintf (d_buf, "%d", new_paneBuffer_id);
      __ctalkSetObjectValueVar (paneBuffer_object, d_buf);
      sprintf (d_buf, "%d", new_paneBackingStore_id);
      __ctalkSetObjectValueVar (paneBackingStore_object, d_buf);
      sprintf (d_buf, "%d", width);
      __ctalkSetObjectValueVar (pane_size_point_x_object, d_buf);
      sprintf (d_buf, "%d", height);
      __ctalkSetObjectValueVar (pane_size_point_y_object, d_buf);
      return SUCCESS;
    } else {
#ifndef WITHOUT_X11LIB_WARNINGS
      printf ("__ctalkX11ResizeWindow: Invalid resize_request return.\n");
      printf ("__ctalkX11ResizeWindow: (To disable these messages, build Ctalk with\n");
      printf ("__ctalkX11ResizeWindow: the --without-x11-warnings option.)\n");
#endif

      if (++n_resize_retries < MAX_RESIZE_RETRIES) {
	/*
	 *  There's a possibility, if there are invalid resource
	 *  id's  in the return buffers, that a valid Pixmap
	 *  buffer can be lost, resulting in a memory 
	 *  leak.  Mostly applies to OS X, which seems 
	 *  to have trouble sometimes detaching shared
	 *  memory correctly.  So we retry the resize anyway, 
	 *  creating completely new buffers.  OS X, at least
	 *  does not seem to require this function to generate
	 *  a synthetic event to indicate that the window 
	 *  must be redrawn completely.
	 */
	__ctalkSetObjectValueVar (paneBuffer_object, "0");
	__ctalkSetObjectValueVar (paneBackingStore_object, "0");
	goto try_resize_request;
      } else {
	return ERROR;
      }
    }
  } else { /*  if (__have_bitmap_buffers (self_object)) */
    sprintf (d_buf, "%d:%d:%d:%d:%d", width, height, 0, 0, 0);
    make_req (shm_mem, PANE_RESIZE_REQUEST,
	      win_id_value->__o_value,
	      gc_value->__o_value, d_buf);
#ifdef GRAPHICS_WRITE_SEND_EVENT
    send_event.xgraphicsexpose.type = GraphicsExpose;
    send_event.xgraphicsexpose.send_event = True;
    send_event.xgraphicsexpose.display = display;
    send_event.xgraphicsexpose.drawable = atoi(win_id_value->__o_value);
    XSendEvent (display, atoi(win_id_value->__o_value),
		False, 0L, &send_event);
#endif
    wait_req (shm_mem);
    return SUCCESS;
  }
}

int __ctalkX11MoveWindow (OBJECT *self, int x, int y) {
  OBJECT *self_object, *win_id_value, *gc_value;
  char d_buf[MAXLABEL];
#ifdef GRAPHICS_WRITE_SEND_EVENT
  XEvent send_event;
#endif
  self_object = (IS_VALUE_INSTANCE_VAR (self) ? self->__o_p_obj : self);
  win_id_value = __x11_pane_win_id_value_object (self_object);
  gc_value = __x11_pane_win_gc_value_object (self_object);
  sprintf (d_buf, "%d:%d", x, y);
  make_req (shm_mem, PANE_MOVE_REQUEST,
	    win_id_value->__o_value,
	    gc_value->__o_value, d_buf);
#ifdef GRAPHICS_WRITE_SEND_EVENT
  send_event.xgraphicsexpose.type = GraphicsExpose;
  send_event.xgraphicsexpose.send_event = True;
  send_event.xgraphicsexpose.display = display;
  send_event.xgraphicsexpose.drawable = atoi(win_id_value->__o_value);
  XSendEvent (display, atoi(win_id_value->__o_value),
 	      False, 0L, &send_event);
#endif
  wait_req (shm_mem);
  return SUCCESS;
}

/*
 *  The I/O error handler lets us shutdown correctly if 
 *  WM_DELETE_WINDOW happens to kill the app (the parent
 *  process of __ctalkX11InputClient) when it deletes the 
 *  window.
 */
int __ctalkX11IOErrorHandler (Display *d) {
  int pending;

#if X11LIB_PRINT_COPIOUS_DEBUG_INFO
  fprintf (stderr, "Ctalk x11lib.c. PID %d\n", (int)getpid ());
  fprintf (stderr, "Display = %p. Pending = %d\n", d, pending);
#endif
  if (d) {
    while ((pending = XPending (d)) > 0)
      XSync (d, TRUE);
    XCloseDisplay (d);
  }
  __ctalkErrorExit ();
#if X11LIB_PRINT_COPIOUS_DEBUG_INFO
  fprintf (stderr, "Display = %p. Pending = %d\n", d, pending);
  (*default_io_handler)(d);
#endif
  return SUCCESS;
}

int __ctalkX11ErrorHandler (Display *d, XErrorEvent *e) {
#ifndef WITHOUT_X_PROTOCOL_ERRORS  
  char error_text[MAXMSG];
  printf ("__ctalkX11ErrorHandler: X11 Protocol Error %d\n", 
	  (int)e -> error_code);
  XGetErrorText (d, (int)e -> error_code, error_text, MAXMSG);
  printf ("__ctalkX11ErrorHandler: %s\n", error_text);
  printf ("__ctalkX11ErrorHandler: (To disable this warning, use the\n");
  printf ("__ctalkX11ErrorHandler: --without-x-protocol-errors option when building Ctalk.)\n");
#endif /* WITHOUT_X_PROTOCOL_ERRORS */
  return SUCCESS;
}

#endif /* !X11LIB_FRAME */

#else /* #if defined (DJGPP) || defined (WITHOUT_X11) */

static void __x_support_error (void) {
  _error ("__ctalkX11 ... () functions require X Window System support.\n");
}

int __ctalkX11InputClient (OBJECT *streamobject, int parent_fd, int mem_handle, int main_win_id) {__x_support_error (); return ERROR;}
int __ctalkCreateX11MainWindow (OBJECT *self_object) {__x_support_error ();return ERROR;}
int __ctalkCreateX11SubWindow (OBJECT *parent, OBJECT *self) {__x_support_error ();return ERROR;}
int __ctalkMapX11Window (OBJECT *self_object) {__x_support_error ();return ERROR;}
int __ctalkRaiseX11Window (OBJECT *self_object) {__x_support_error();return ERROR;}
int __ctalkX11SetWMNameProp (OBJECT *self_object, char *title) {__x_support_error();return ERROR;}
int __ctalkX11UseFont (OBJECT *self) {__x_support_error(); return ERROR;}
int __ctalkX11ResizeWindow (OBJECT *self, int width, int height) {__x_support_error(); return ERROR;}
int __ctalkX11MoveWindow (OBJECT *self, int x, int y) {__x_support_error(); return ERROR;}
#endif  /* #if defined (DJGPP) || defined (WITHOUT_X11) */
